@@ -21,13 +21,18 @@ def async_signal_handler():
 
     Disconnect all active async clients.
     """
-    async def _handler():
-        asyncio.get_event_loop().stop()
+    async def _handler():  # pragma: no cover
         for c in client.connected_clients[:]:
             if c.is_asyncio_based():
                 await c.disconnect()
-        else:  # pragma: no cover
-            pass
+
+        # cancel all running tasks
+        tasks = [task for task in asyncio.all_tasks() if task is not
+                 asyncio.current_task()]
+        for task in tasks:
+            task.cancel()
+        await asyncio.gather(*tasks, return_exceptions=True)
+        asyncio.get_event_loop().stop()
 
     asyncio.ensure_future(_handler())
 
@@ -62,7 +67,11 @@ class AsyncClient(client.Client):
                           leave interrupt handling to the calling application.
                           Interrupt handling can only be enabled when the
                           client instance is created in the main thread.
+    :param websocket_extra_options: Dictionary containing additional keyword
+                                    arguments passed to
+                                    ``aiohttp.ws_connect()``.
     """
+
     def is_asyncio_based(self):
         return True
 
@@ -95,9 +104,9 @@ class AsyncClient(client.Client):
             try:
                 asyncio.get_event_loop().add_signal_handler(
                     signal.SIGINT, async_signal_handler)
-                async_signal_handler_set = True
             except NotImplementedError:  # pragma: no cover
                 self.logger.warning('Signal handler is unsupported')
+        async_signal_handler_set = True
 
         if self.state != 'disconnected':
             raise ValueError('Client is not in a disconnected state')
@@ -126,9 +135,9 @@ class AsyncClient(client.Client):
             await self.read_loop_task
 
     async def send(self, data):
-        """Send a message to a client.
+        """Send a message to the server.
 
-        :param data: The data to send to the client. Data can be of type
+        :param data: The data to send to the server. Data can be of type
                      ``str``, ``bytes``, ``list`` or ``dict``. If a ``list``
                      or ``dict``, the data will be serialized as JSON.
 
@@ -158,9 +167,7 @@ class AsyncClient(client.Client):
                 client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
-            if self.http and not self.http.closed:  # pragma: no cover
-                await self.http.close()
-        self._reset()
+        await self._reset()
 
     def start_background_task(self, target, *args, **kwargs):
         """Start a background task.
@@ -193,6 +200,12 @@ class AsyncClient(client.Client):
         """Create an event object."""
         return asyncio.Event()
 
+    async def _reset(self):
+        super()._reset()
+        if not self.external_http:  # pragma: no cover
+            if self.http and not self.http.closed:
+                await self.http.close()
+
     def __del__(self):  # pragma: no cover
         # try to close the aiohttp session if it is still open
         if self.http and not self.http.closed:
@@ -217,11 +230,11 @@ class AsyncClient(client.Client):
             'GET', self.base_url + self._get_url_timestamp(), headers=headers,
             timeout=self.request_timeout)
         if r is None or isinstance(r, str):
-            self._reset()
+            await self._reset()
             raise exceptions.ConnectionError(
                 r or 'Connection refused by the server')
         if r.status < 200 or r.status >= 300:
-            self._reset()
+            await self._reset()
             try:
                 arg = await r.json()
             except aiohttp.ClientError:
@@ -297,18 +310,22 @@ class AsyncClient(client.Client):
                 break
         self.http.cookie_jar.update_cookies(cookies)
 
+        extra_options = {'timeout': self.request_timeout}
+        if not self.ssl_verify:
+            ssl_context = ssl.create_default_context()
+            ssl_context.check_hostname = False
+            ssl_context.verify_mode = ssl.CERT_NONE
+            extra_options['ssl'] = ssl_context
+
+        # combine internally generated options with the ones supplied by the
+        # caller. The caller's options take precedence.
+        headers.update(self.websocket_extra_options.pop('headers', {}))
+        extra_options['headers'] = headers
+        extra_options.update(self.websocket_extra_options)
+
         try:
-            if not self.ssl_verify:
-                ssl_context = ssl.create_default_context()
-                ssl_context.check_hostname = False
-                ssl_context.verify_mode = ssl.CERT_NONE
-                ws = await self.http.ws_connect(
-                    websocket_url + self._get_url_timestamp(),
-                    headers=headers, ssl=ssl_context)
-            else:
-                ws = await self.http.ws_connect(
-                    websocket_url + self._get_url_timestamp(),
-                    headers=headers)
+            ws = await self.http.ws_connect(
+                websocket_url + self._get_url_timestamp(), **extra_options)
         except (aiohttp.client_exceptions.WSServerHandshakeError,
                 aiohttp.client_exceptions.ServerConnectionError,
                 aiohttp.client_exceptions.ClientConnectionError):
@@ -501,7 +518,7 @@ class AsyncClient(client.Client):
                 client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
-            self._reset()
+            await self._reset()
         self.logger.info('Exiting read loop task')
 
     async def _read_loop_websocket(self):
@@ -512,10 +529,13 @@ class AsyncClient(client.Client):
                 p = await asyncio.wait_for(
                     self.ws.receive(),
                     timeout=self.ping_interval + self.ping_timeout)
-                p = p.data
-                if p is None:  # pragma: no cover
+                if not isinstance(p.data, (str, bytes)):  # pragma: no cover
+                    self.logger.warning(
+                        'Server sent unexpected packet %s data %s, aborting',
+                        str(p.type), str(p.data))
                     await self.queue.put(None)
                     break  # the connection is broken
+                p = p.data
             except asyncio.TimeoutError:
                 self.logger.warning(
                     'Server has stopped communicating, aborting')
@@ -549,7 +569,7 @@ class AsyncClient(client.Client):
                 client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
-            self._reset()
+            await self._reset()
         self.logger.info('Exiting read loop task')
 
     async def _write_loop(self):
@@ -564,10 +584,11 @@ class AsyncClient(client.Client):
             packets = None
             try:
                 packets = [await asyncio.wait_for(self.queue.get(), timeout)]
-            except (self.queue.Empty, asyncio.TimeoutError,
-                    asyncio.CancelledError):
+            except (self.queue.Empty, asyncio.TimeoutError):
                 self.logger.error('packet queue is empty, aborting')
                 break
+            except asyncio.CancelledError:  # pragma: no cover
+                break
             if packets == [None]:
                 self.queue.task_done()
                 packets = []
@@ -599,7 +620,7 @@ class AsyncClient(client.Client):
                 if r.status < 200 or r.status >= 300:
                     self.logger.warning('Unexpected status code %s in server '
                                         'response, aborting', r.status)
-                    self._reset()
+                    await self._reset()
                     break
             else:
                 # websocket
